from typing import Literal, Optional

from pydantic import BaseModel, ConfigDict, Field

from emp_agents.models.shared.message import Message, SystemMessage
from emp_agents.models.shared.tools import GenericTool
from emp_agents.types import ModelType, Role


class Request(BaseModel):
    """
    https://platform.openai.com/docs/api-reference/chat/create
    """

    model_config = ConfigDict(populate_by_name=True)

    model: ModelType
    max_tokens: Optional[int] = Field(default=1_000, lt=16_384, gt=0)
    temperature: Optional[float] = Field(default=None, ge=0, le=2.0)
    tool_choice: Literal["none", "required", "auto", None] = Field(default=None)
    tools: Optional[list[GenericTool]] = None
    response_format: type[BaseModel] | None = None

    system: str | None = None  # anthropic field
    messages: list[Message]

    frequency_penalty: Optional[float] = Field(default=None, ge=-2.0, le=2.0)  # openai
    presence_penalty: Optional[float] = Field(default=None, ge=-2.0, le=2.0)  # openai
    num_responses: Optional[int] = Field(
        default=None, serialization_alias="n"
    )  # openai
    top_p: Optional[int] = Field(default=None)  # openai

    def model_dump(self, *, exclude_none=True, by_alias=True, **kwargs):
        return super().model_dump(
            exclude_none=exclude_none, by_alias=by_alias, **kwargs
        )

    def to_anthropic(self):
        exclude = ["frequency_penalty", "presence_penalty", "num_responses", "n"]
        result = self.model_dump(exclude_none=True)
        result["tools"] = [t.to_anthropic() for t in self.tools] if self.tools else []
        if "tool_choice" in result:
            result["tool_choice"] = {"type": result["tool_choice"]}
        for field in exclude:
            if field in result:
                del result[field]
        messages = [message for message in self.messages if message.role != Role.system]
        system_messages = [
            message for message in self.messages if message.role == Role.system
        ]
        result["system"] = result.get("system", "") + str(
            "\n".join([m.content for m in system_messages])
        )
        result["messages"] = [m.model_dump(exclude_none=True) for m in messages]
        return result

    def to_openai(self):
        exclude = ["system"]
        result = self.model_dump(exclude_none=True)
        if self.system:
            messages = [SystemMessage(content=self.system)] + self.messages
        else:
            messages = self.messages

        # Function to recursively set 'additionalProperties': False
        def set_additional_properties_false(schema):
            if isinstance(schema, dict):
                if schema.get("type") == "object":
                    schema["additionalProperties"] = False
                for key, value in schema.items():
                    set_additional_properties_false(value)
            elif isinstance(schema, list):
                for item in schema:
                    set_additional_properties_false(item)

        if "response_format" in result:
            model_schema = self.response_format.model_json_schema()
            set_additional_properties_false(model_schema)
            del result["response_format"]
            result["response_format"] = {
                "type": "json_schema",
                "json_schema": {
                    "name": self.response_format.__name__,
                    "description": "response format",
                    "strict": True,
                    "schema": {
                        "type": "object",
                        "additionalProperties": False,
                        **model_schema,
                    },
                },
            }
        result["messages"] = [m.model_dump() for m in messages]
        result["tools"] = (
            [t.to_openai().model_dump(exclude_none=True) for t in self.tools]
            if self.tools
            else None
        )

        for field in exclude:
            if field in result:
                del result[field]
        return result



from enum import StrEnum


class OpenAIModelType(StrEnum):
    gpt3_5 = "gpt-3.5-turbo-0125"
    gpt3_5_turbo = "gpt-3.5-turbo"
    gpt4 = "gpt-4"
    gpt4_turbo = "gpt-4-turbo"
    gpt4o_mini = "gpt-4o-mini"  # 128_000 tokens
    gpt4o = "gpt-4o"
    gpt_o1_mini = "o1-mini"
    gpt_o1_preview = "o1-preview"


class AnthropicModelType(StrEnum):
    claude_3_5_sonnet = "claude-3-5-sonnet-20240620"
    claude_3_opus = "claude-3-opus-20240229"
    claude_3_sonnet = "claude-3-sonnet-20240229"
    claude_3_5_haiku = "claude-3-5-haiku-20241022"
    claude_3_haiku = "claude-3-haiku-20240307"
    claude_2_1 = "claude-2.1"
    claude_2_0 = "claude-2.0"
    claude_instant_1_2 = "claude-instant-1.2"


class Role(StrEnum):
    system = "system"
    user = "user"
    assistant = "assistant"
    tool = "tool"


ModelType = AnthropicModelType | OpenAIModelType



import os
from typing import ClassVar

from anthropic import AsyncAnthropic as Anthropic
from pydantic import BaseModel, ConfigDict, Field, PrivateAttr

from emp_agents.models.anthropic.response import Response
from emp_agents.models.shared import Request
from emp_agents.types import TCompletionAgent


class AnthropicBase(BaseModel, TCompletionAgent[Response]):
    URL: ClassVar[str] = "https://api.openai.com/v1/chat/completions"

    anthropic_api_key: str = Field(
        default_factory=lambda: os.environ["ANTHROPIC_API_KEY"]
    )
    _client: Anthropic = PrivateAttr()

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def model_post_init(self, __context) -> None:
        self._client = Anthropic(api_key=self.anthropic_api_key)
        return super().model_post_init(__context)

    @property
    def headers(self):
        return {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.anthropic_api_key}",
        }

    async def completion(self, request: Request) -> Response:
        message = await self._client.messages.create(**request.to_anthropic())
        try:
            return Response(**message.model_dump())
        except Exception as e:
            raise e



from typing import ClassVar

import httpx
from pydantic import BaseModel

from emp_agents.models.openai.request import Message, Tool
from emp_agents.models.openai.response import Response
from emp_agents.models.openai.tool import Function, Parameters, Property
from emp_agents.models.openai.types import Classification
from emp_agents.models.shared import Request
from emp_agents.types import TCompletionAgent


class OpenAIBase(BaseModel, TCompletionAgent[Response]):
    URL: ClassVar[str] = "https://api.openai.com/v1/chat/completions"
    openai_api_key: str

    @property
    def headers(self):
        return {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.openai_api_key}",
        }

    async def completion(self, request: Request) -> Response:
        async with httpx.AsyncClient(headers=self.headers) as client:
            response = await client.post(
                self.URL, json=request.to_openai(), timeout=None
            )
        if response.status_code >= 400:
            raise ValueError(response.json())
        return Response(**response.json())


__all__ = [
    "Classification",
    "Message",
    "OpenAIBase",
    "Request",
    "Response",
    "Tool",
    "Function",
    "Property",
    "Parameters",
]



from enum import StrEnum

from pydantic import BaseModel


class Classification(BaseModel):
    name: str
    description: str


class FinishReason(StrEnum):
    stop = "stop"
    length = "length"
    function_call = "function_call"
    content_filter = "content_filter"
    null = "null"
    tool_calls = "tool_calls"



from . import tools
from .agents import AgentBase
from .models import AnthropicBase, GenericTool, Message, OpenAIBase, Property, Request
from .types import AnthropicModelType, OpenAIModelType, Role

__all__ = [
    "AgentBase",
    "AnthropicBase",
    "AnthropicModelType",
    "GenericTool",
    "Message",
    "OpenAIBase",
    "OpenAIModelType",
    "Property",
    "Request",
    "Role",
    "tools",
]



from emp_agents.types.completion import TCompletionAgent

from .enums import AnthropicModelType, ModelType, OpenAIModelType, Role

__all__ = [
    "TCompletionAgent",
    "AnthropicModelType",
    "OpenAIModelType",
    "Role",
    "ModelType",
]



from emp_agents.models.shared.message import AssistantMessage, ToolMessage, UserMessage

QuestionAnswer = tuple[UserMessage, AssistantMessage]
TooledMessageSequence = tuple[UserMessage, AssistantMessage, list[ToolMessage]]

Example = QuestionAnswer | TooledMessageSequence
